# SPDX-FileCopyrightText: ☭ Emery Hemingway
# SPDX-License-Identifier: Unlicense

import std/[deques, streams]
import bumpy, pixie, sdl2

type
  Rect = bumpy.Rect

  Pane* = ref object
    image: pixie.Image
    spans: seq[Span]
    arrangement: Arrangement

  Well* = ref object
    panes: Deque[Pane]
    dimensions: Vec2

using
  pane: Pane
  well: Well

proc newWell*(width, height: int): Well =
  result = Well(dimensions: vec2(float width, float height))
  result.panes.addFirst Pane()

proc margins*(well): Vec2 = well.dimensions / 9.0

proc bounds*(well): Vec2 = well.margins * 7.0

proc rect*(well): Rect =
  result.wh = well.dimensions

proc rectAt(well; offset: int): Rect =
  if offset < well.panes.len:
    let n = succ offset
    result.w = well.dimensions.x / float(2 * n)
    result.h = well.dimensions.y / float(2 * n)
    case offset and 3
    of 0: (result.x, result.y) = (result.w, result.h)
    of 1: result.y = result.h
    of 2: result.x = result.w
    else: discard

proc place*(well; offset, width, height: int): (Pane, Rect) =
  ## Return the `Pane` at `offset` from the top of `well` or `nil` if
  ## the offset is too deep. The position and size of the `Pane` is
  ## returned as well.
  if offset < well.panes.len:
    result[0] = well.panes[offset]
    result[1] = well.rectAt(offset)

proc append*(well; text: string; font: Font) =
  assert well.panes.len > 0
  let span = newSpan(text, font)
  while true:
    let pane = well.panes[0]
    pane.spans.add(span)
    var
      arrangement = typeset(pane.spans, well.bounds)
      bounds = layoutBounds arrangement
    # assert bounds.x <= well.bounds.x
    if bounds.y <= well.bounds.y:
      doAssert pane.spans.len > 0, "text does not find on a single pane - " & $bounds & $well.bounds
      pane.arrangement = arrangement
      break
    else:
      well.panes.addFirst Pane()

proc append*(well; stream: Stream; font: Font) =
  var line: string
  while readLine(stream, line):
    append(well, line, font)

#[
proc render*(well; index: Natural; scale: float): Image =
  if index < well.panes.len:
    let dimensions = well.dimensions * scale
    if dimensions.x > 8 and dimensions.y > 8:
      result = newImage(int dimensions.x, int dimensions.y)
      fill(result, rgba(255, 255, 255, 255))
      fillText(well.panes[index].arrangment, translate(well.margin))
]#

iterator intersectingPanes*(well; rect: Rect): (int, Rect) =
  var i = 0
  while i < well.panes.len:
    let bounds = well.rectAt(i)
    if (i and 3) == 3:
      let quad = rect(vec2(0,0), bounds.wh * 2.0)
      if not rect.overlaps quad:
        # all further panes are non-intersecting
        break
    if rect.overlaps bounds:
      yield (i, bounds)
    inc(i)

const
  amask = uint32 0xff000000
  rmask = uint32 0x000000ff
  gmask = uint32 0x0000ff00
  bmask = uint32 0x00ff0000

proc texture*(well; index: int; renderer: RendererPtr; wh: Vec2): TexturePtr =
  # TODO:­caching
  if index >= well.panes.len or wh.x < 4 or wh.y < 4: return nil
  let pane = well.panes[index]
  let
    image = newImage(wh.x.int, wh.y.int)
    zoom = wh / well.dimensions
    margin = wh / 9.0
  if not pane.arrangement.isNil:
    image.fillText(pane.arrangement, scale(zoom) * translate(margin))
  var surface = createRGBSurfaceFrom(
      image.data[0].addr, wh.x.cint, wh.y.cint, cint 32, 4 * wh.x.cint,
      rmask, gmask, bmask, amask,
    )
  result = createTextureFromSurface(renderer, surface)
  destroy(surface)

